package crowdUser;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.net.HttpURLConnection;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLEncoder;
import java.util.*;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

import myDataBases.MySQLbase;
import myDataBases.MediaBase;


/**
 * Serves as an intermediary between the MySQLbase database in which HIT results are stored and CrowdFlower, our
 * <a href="https://secure.wikimedia.org/wikipedia/en/wiki/Crowdsourcing">crowdsourcing</a> provider.
 * 
 * <p>This class performs several operations : after retrieving unknown comparisons, it submits them to the crowd using 
 * CrowdFlower's "self service". When the HITs are finished, it downloads the results from the Antechamber (see the
 * <a href="http://ns204207.ovh.net/videosense/">online documentation</a>), saves them, gives them to its {@link Refiner}
 * instance (<tt>refiner</tt>) to parse them and sends them to <tt>dbSQL</tt>, a {@link myDataBases.MySQLbase} instance
 * in which they are stored.</p>
 * <p>Once this is done, it uses <tt>refiner</tt> to reduce the noise in the hard results and thus to obtain refined 
 * results (more details in the {@link Refiner} documentation). These are then sent to <tt>dbSQL</tt>.</p>
 * 
 * @author	Leo Perrin (perrin.leo@gmail.com)
 */
public class CrowdManager {

  //
  // Fields
  //

  /**
   * The ID of the media alleged to be "greater". They are to be compared with those of the "smaller" attribute
   * through HITs.
   */
  private List<Integer> greater;
  /**
   * The ID of the media alleged to be "smaller". They are to be compared with those of the "greater" attribute
   * through HITs.
   */
  private List<Integer> smaller;
  /**
   * The axes along which each comparison must be performed.
   */
  private List<String> axes;
  /**
   * The number of comparisons performed during the current iteration.
   */
  private Integer comparisonsNumber;
  /**
   * The key given by crowdFlower to authenticate yourself. If you have an account (and you definitively have to), it is
   * available <a href="https://crowdflower.com/account">here</a>.
   */
  private String cfKey ;
  /**
   * The identifier of the "original" job, i.e the one from which every characteristics
   * (<a href="https://crowdflower.com/docs/api/webhooks/">webhook</a> address, tags,
   * <a href="https://crowdflower.com/docs/cml">cml</a> configuration...) will be copied in all the jobs created during
   * the current execution of CPS. 
   */
  private Integer originalJobId;
  /**
   * The number of jobs already submitted (a job is a group of HITs)
   */
  private String resultsUrl;
  /**
   * The MediaBase instance in which the names of the media are stored
   */
  private MediaBase dbMedia ;
  /**
   * The MySQLbase instance to which it is connected and in which HIT results must be stored.
   */
  private MySQLbase dbSQL;
  /**
   * The Refiner instance used to treat the result of the HITs.
   */
  private Refiner refiner;

  
  //
  // Constructors
  //
  /**
   * Creates an instance of the CrowdManager class.
   * 
   * Initializes all of its attributes, except the <tt>greater, smaller</tt> and <tt>axes</tt> attributes. Indeed, these
   * are set afterward using affectation like <tt>CrowdManagerInstance.getGreater() = someArrayOfIntegers</tt>.
   *
   * @param		firstJobId The identifier of a job in your CrowdFlower account already configured.
   * @param		key	The key given by CrowdFlower.
   * @param		resultsUrl	The URL where <tt>results$JOB_ID.txt</tt> is, most likely that of the <tt>csv/</tt> folder of
   * the Antechamber.
   * @param		baseSQL	The <tt>MySQLbase</tt> in which the results must be stored.
   * @param		baseDIR	The <tt>MediaBase</tt> in which the name of the media sorted are stored
   * @param		fieldsHardData	The fields of the CSV file returned by the Antechamber containing the HITs' hard results.
   */
  public CrowdManager (Integer firstJobId, String key, String resultsUrl, MySQLbase baseSQL,
		  					MediaBase baseDIR, List<String> fieldsHardData)
  {
	  this.comparisonsNumber = 0;
	  this.resultsUrl = resultsUrl;
	  this.greater = new ArrayList<Integer>();
	  this.smaller = new ArrayList<Integer>();
	  this.axes = new ArrayList<String>();
	  this.originalJobId = firstJobId;
	  this.cfKey = key;
	  this.dbSQL = baseSQL;
	  this.dbMedia = baseDIR ;
	  this.refiner = new Refiner(fieldsHardData);
  };
  
  //
  // Methods
  //
  
  /**
   * Creates a new job at CrowdFlower that is a copy of the "original" one (the one with the <tt>originalJobId</tt>
   * identifier) and returns its identifier.
   * 
   * <p>Creates a copy of of the original job using a http POST to the address given in Crowdflower's
   * <a href="https://crowdflower.com/docs/api/jobs#copy">documentation</a>. When the POST is done, CrowdFlower returns a
   * JSON formatted string containing, among other things, the identifier of the job.</p>
   * 
   * @return	The identifier of the new job.
   */
  public Integer createNewJob()
  {
	  Integer id = 0;

	  String postUrl = "http://api.crowdflower.com/v1/jobs/" + this.originalJobId
	  		+ "/copy.json?key=" + this.cfKey ;
	  // posting the request
	  try
	  {
	  	  URL url = new URL(postUrl);
	  	  HttpURLConnection conn = (HttpURLConnection) url.openConnection();
	  	  // Building an HTTP request
	  	  conn.setDoOutput(true);
	  	  conn.setRequestMethod("POST");
	  	  conn.connect();
	  	  // getting the answer from CrowdFlower and reading the "id" field
	  	  BufferedReader rd = new BufferedReader(new InputStreamReader(conn.getInputStream()));
	  	  String line, idString = "";
	  	  while ((line = rd.readLine()) != null)
	  		  if (line.contains('"'+"id"+'"'))
	  			  // we go to the index of ' "id": ' and read the number. In order no to take '"id":', 5 is added.
	  			  for (int i = line.lastIndexOf('"' +"id"+'"')+5; i<line.length(); i++)
	  				  if ( line.charAt(i) == ',' )
	  					  break;
	  				  else
	  					  idString += line.charAt(i) ;
	  	  id = Integer.parseInt(idString);
	  	  rd.close();
	  } catch (Exception e) { e.printStackTrace(); }

	  return id;
  }


  
  /**
   * Uploads at CrowdFlower the csv file corresponding to the comparisons required.
   * 
   * <p>The <a href="https://secure.wikimedia.org/wikipedia/en/wiki/Comma-separated_values">csv</a> (Comma Separated File)
   * file is first generated. The fields used are NOT set using the config.xml, they are hard coded. Indeed, CrowdFlower
   * needs nothing else than the axis along which the comparison is performed, the identifier of the media (so they can be
   * sent back), their names (so they can be displayed) and a text associated with each one (for example to respect a 
   * CC-BY license).</p>
   * <p>Once generated, it is sent using a PUT http request containing the csv file, the identifier of the job
   * (<tt>jobNUmber</tt>) and the key given by CrowdFlower (the <tt>key</tt> parameter).</p> 
   * 
   * @param		jobId	The identifier of the job that will receive the data contained in the csv file. 
   */
  public void uploadHIT(Integer jobId)
  {
	  this.comparisonsNumber += this.greater.size() ;
	  String post = "";
	  // writing the .csv file and the post String
	  try{
		  BufferedWriter fichier = new BufferedWriter
		  			(new FileWriter("../data/HITdata/" + jobId + ".csv"));
		  String line = "axis,idMedia1,idMedia2,urlMedia1,urlMedia2,miscData1,miscData2";
		  fichier.write(line);
	      fichier.newLine();
	      post += line ;
		  for (int i=0; i<this.greater.size(); i++)
		  {
			  line = this.axes.get(i) + "," + this.greater.get(i) + "," + this.smaller.get(i) +
		     		"," + this.dbMedia.getMedia(this.greater.get(i)) +
		     		"," + this.dbMedia.getMedia(this.smaller.get(i)) +
		     		",\"" + this.dbMedia.getContent().get(this.greater.get(i)).get("miscData") +
		     		"\",\"" + this.dbMedia.getContent().get(this.smaller.get(i)).get("miscData") + "\"";
			  fichier.write(line);
		      fichier.newLine();
		      post += "\n" + line ;
		  }
	      fichier.close();
	      
	  }
	  catch (Exception e) { e.printStackTrace(); }
	  
	  // posting the file

	  try
	  {
		  URL url = new URL("http://api.crowdflower.com/v1/jobs/" + jobId +
			  	"/upload.json?key=" + this.cfKey);
		  HttpURLConnection conn = (HttpURLConnection) url.openConnection();
		  // Building an HTTP request
		  conn.setDoOutput(true);
		  conn.setRequestMethod("PUT");
		  conn.setRequestProperty("Content-Type", "text/csv");
		  OutputStreamWriter osw = new OutputStreamWriter(conn.getOutputStream());
		  osw.write(post);
		  osw.flush();
		  osw.close();
		  
		  // getting the answer from CrowdFlower
		  BufferedReader rd = new BufferedReader(new InputStreamReader(conn.getInputStream()));
		  String line;
		  while ((line = rd.readLine()) != null) {
		  	System.out.println(line);
		  }
		  
		  rd.close();
	  } catch (Exception e) { e.printStackTrace(); }

  }


  /**
   * Downloads results from the URL specified in the <tt>resultsUrl</tt> attribute and returns the path to it minus its 
   * extension (<em>i.e</em> <tt>'../data/HITresults/results42'</tt> for the results stored in the 
   * <tt>'../data/HITresults/results42<strong>.csv</strong>'</tt> file).
   * 
   * <p>First, you have to make sure the results are at the given address: was the correct
   * <a href="https://crowdflower.com/docs/api/webhooks/">webhook</a> sent at the Antechamber ? Is the processing of the
   * raw results finished? An email is sent when both of these operations are finisehded.</p>
   * <p>If so, use this method to download the <tt>results$JOB_ID.txt</tt> file containing the answers of the
   * workers and to save it as <tt>../data/results/results$JOB_ID.csv</tt>.</p>
   * 
   * @param			jobId	The identifier of the job used to obtain comparisons. It is given by CrowdFlower's web
   * application when a new job is created as well as the <tt>createNewJob()</tt> method.
   * @return		The path to the csv file containing the results of the job.
   */
  public String getHITresults(Integer jobId)
  {
	  try
   	  {
		  // creating connection
		  String urlString = this.resultsUrl + "csv/" + jobId + ".txt";
		  System.out.println(urlString);
		  URL url = new URL(urlString);
		  System.out.println("Opening connection to " + urlString + "...");
		  // reading input
		  InputStream is = url.openStream();
		  System.out.flush();
		  // creating results$JOB_ID.csv file
		  FileOutputStream fos=null;
		  fos = new FileOutputStream("../data/HITresults/results" + jobId + ".csv");
		  // writing to file
		  int oneChar, count=0;
		  while ((oneChar=is.read()) != -1)
		  {
			  fos.write(oneChar);
			  count++;
		  }
		  // close everything
		  is.close();
		  fos.close();
		  System.out.println("csv results file downloaded, " + count + " byte(s) copied");
   	  }
      catch (Exception e) { e.printStackTrace(); }
      // return the name of the file minus ".csv"
	  return "results" + jobId ;
  }
  
  
  /**
   * Adds the HIT results contained in the csv file in <tt>path</tt> to the <tt>dbSQL</tt> {@link myDataBases.MySQLbase}
   * database.
   * 
   * <p>The hard results are sent to <tt>refiner</tt> in order it to parse them. They are then sent to <tt>dbSQL</tt> to be
   * stored. At this point, <tt>dbSQL</tt> contains all of the results obtained.</p>
   * <p>After that, results are refined by <tt>refiner</tt> in order to reduce the noise, see the 
   * <tt>Refiner.getanotherlabel</tt> method for more details on how it is done. Again, when this is over, refined results
   * are stored in <tt>dbSQL</tt>.</p>
   * 
   * @param		path The path to the csv file containing the HIT results
   * 
   * @see Refiner
   */
  public void readHITresults( String path )
  {
	  List<String[]> entries ;
	  // retrieve hard data from the csv file
	  this.refiner.csv2List(path) ;
	  entries = this.refiner.getData() ;
	  this.dbSQL.insertHardDataInSQL(entries);
	  // retrieve good data from the refiner
	  this.refiner.getanotherlabel() ;
	  entries = this.refiner.getData() ;
	  this.dbSQL.insertResultsInSQL(entries,this.dbMedia) ;
  }


  /**
   * Asks the <tt>Antechamber</tt> to send an email signaling the end of the Splitsort.
   * 
   * <p>Uses an http <tt>POST</tt> powered by java.net to the correct URL to ask the <tt>Antechamber</tt> to send an e-mail
   * containing the identifier of this last job, information on how to retrieve the results and the content of the 
   * <tt>results</tt> String as an attachment.</p>
   * 
   * @param		antechamberUrl	The url to which the information must be posted
   * @param 	results Content to be sent as an attachment file along with the e-mail. The name it will have is set in
   * the Antechamber.
   */
  // Yes, little friend / This is the end / My only friend / the end / Of everything that stands / the end
  // I'll never look into your eyes / again...
  public static void thisIsTheEnd(String antechamberUrl, String results)
  {
	  try
	  {
		  URL url = new URL(antechamberUrl);
		  HttpURLConnection conn = (HttpURLConnection) url.openConnection();
		  // Building an HTTP request
		  conn.setDoOutput(true);
		  conn.setRequestMethod("POST");
		  OutputStreamWriter osw = new OutputStreamWriter(conn.getOutputStream());
		  String data = URLEncoder.encode("signal", "UTF-8") + "=" + URLEncoder.encode("sort_finished","UTF-8")
		  	+ "&" + URLEncoder.encode("payload", "UTF-8") + "=" + URLEncoder.encode(results, "UTF-8");
		  osw.write(data);
		  osw.flush();
		  osw.close();
		  // getting the answer from CrowdFlower
		  BufferedReader rd = new BufferedReader(new InputStreamReader(conn.getInputStream()));
		  String line;
		  while ((line = rd.readLine()) != null) {
		  	System.out.println(line);
		  }
		  rd.close();
	  } catch (Exception e) { e.printStackTrace(); }
  }
  
  //
  // Accessor methods
  //

  /**
   * Returns the value of greater.
   * @return greater
   */
  public List<Integer> getGreater()
  { return this.greater ; }

  
  /**
   * Returns the value of smaller.
   * @return smaller
   */
  public List<Integer> getSmaller()
  { return this.smaller; }

  
  /**
   * Returns the value of axes.
   * @return axes
   */
  public List<String> getAxes()
  { return this.axes; }
  
  
  /**
   * Returns the value of comparisonsNumber.
   * @return comparisonsNumber
   */
  public Integer getComparisonsNumber()
  { return this.comparisonsNumber; }


}
