package crowdUser;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.Statement;
import java.util.*;

import ipeirotis.gal.scripts.DawidSkene;
import ipeirotis.gal.scripts.Labeling;



/**
 * Parses csv formatted raw results and allows their amelioration by
 * <a href="http://code.google.com/p/get-another-label/"><tt>get another label</tt></a>, a program designed by P
 * Ipeirotis (a researcher in computer science specialized in crowdsourcing).
 * 
 * <p>In doing so, the noise is reduced and the comparisons are more accurate, thus assuring a better sort.</p>
 * <p>More generally, this class parses results in order them to be easily used by a CrowdManager instance, for example
 * by transforming a csv file into a list of String.</p>
 * 
 * @author	Leo Perrin (perrin.leo@gmail.com)
 */
public class Refiner {

  //
  // Fields
  //

  /**
   * The labels of the columns in the csv file.
   */
  private String[] columnsCSV;
  /**
   * The column in which the axis considered for the comparison is stored.
   */
  private Integer columnAxis;
  /**
   * The column in which the identifier of the first media is.
   */
  private Integer columnMedia1;
  /**
   * The column in which the identifier of the second media is.
   */
  private Integer columnMedia2;
  /**
   * The column in which the identifier of the worker who performed the comparison.
   */
  private Integer columnWorkerId;
  /**
   * The column in which is stored the result of the question asked to the worker.
   */
  private Integer columnAnswer;
  /**
   * A Map<String,Integer> containing the score of a comparison : for each hit, if the worker said the right member was
   * greater, score+=1 ; else score += -1. Each comparisons is identified by its signature.
   */
  private Map<String,Integer> compScores ;
  /**
   * A List of array of string containing the data currently studied : hard data or treated one, depending on the moment.
   */
  private List<String[]> data;

  
  //
  // Constructors
  //
  /**
   * Creates a Refiner instance and sets the fields considered in the
   * <a href="https://secure.wikimedia.org/wikipedia/en/wiki/Comma-separated_values"> formatted results</a>.
   * 
   * <p>It will use the fields contained in the 'fields' parameter to initialize the <tt>columnsCSV</tt> attribute as well
   * as all those corresponding to a column number (i.e : <tt>column[Axis | Media[1|2] | Answer | WorkerId]</tt>).</p>
   * 
   * @param		fields 		A list containing the fields of the CSV file containing the HIT results. 
   */
  public Refiner (List<String> fields)
  {
	  this.data = new ArrayList <String[]> ();
	  getFields(fields);
  }
  
  //
  // Methods
  //
  
  
  /**
   * Returns the "signature" of a comparison, a String identifying it without ambiguity.
   * 
   * <p>The signature of a comparison is defined as follow : if a comparisons has been performed between <tt>idMedia1</tt>
   * and <tt>idMedia2</tt> along the <tt>axis</tt> axis, its signature is : axis__idMedia1__idMedia2. It is the key
   * under which the result of the comparison is stored in the <tt>compScores</tt> attribute.</p>
   * 
   * @param		axis		the axis along which the studied comparison is performed
   * @param		idMedia1	the identifier of the first media
   * @param		idMedia2	the identifier of the second media
   * 
   * @return	sig, a String : the signature of a comparison.
   *
   * @see #compScores
   */
  private String signature(String axis, String idMedia1, String idMedia2)
  {
	  String sig =  axis + "__" + idMedia1 + "__" + idMedia2;
	  return sig;
  }
  

  /**
   * Initializes <tt>columnsCSV</tt> and <tt>column*</tt> using the content of the 'csvFields' parameter.
   * 
   * <p>For each String contained in <tt>csvFields</tt>, it is checked whether its value is one of the compulsory ones
   * (i.e, <tt>idMedia1</tt>, <tt>idMedia2</tt>, <tt>WorkerId</tt>, <tt>axis</tt>, and a last much longer one
   * corresponding to the question asked), in which case the column number is saved in the corresponding attribute.</p>
   * <p>Although <tt>MiscData[1/2]</tt> are also compulsory fields, they are not saved here in order to save unnecessary
   * space.</p>
   *  
   * @param		csvFields	A lists containing the fields of the CSV file containing the HIT results.
   */
  private void getFields(List<String> csvFields)
  {
	 this.columnsCSV = csvFields.toArray(new String[0]) ;
	 Integer rank = 0;
	 for (String field : csvFields)
	      {
	    	  if (field.equals("idMedia1") )
	    		  this.columnMedia1 = rank ;
	    	  else if (field.equals("idMedia2") )
	    		  this.columnMedia2 = rank ;
	    	  else if (field.equals("_worker_id") )
	    		  this.columnWorkerId = rank ;
	    	  else if (field.equals("axis") )
	    		  this.columnAxis = rank ;
	    	  else if (field.length() > 25 )
	    		  this.columnAnswer = rank ;
	    	  rank ++ ;
	      }
  }

  
 
  /**
   * Puts the data contained in a csv file in the <tt>data</tt> attribute.
   * 
   * <p>The <tt>data</tt> attribute is a List of Arrays of String. Each array is made of the content of one row of the
   * csv file, each one of its case being the content of a column in this line.</p>
   * 
   * @param        csvFile The path to the csv file within the <tt>'../data/HITresults/'</tt> directory, minus the
   * extension (for instance, <tt>results$JOB_ID</tt> instead of <tt>results$JOB_ID.csv</tt>). 
   */
  public void csv2List( String csvFile )
  {

      this.data.clear();
	  try
	  {
		  // connecting to the .csv "database"
		  Class.forName("org.relique.jdbc.csv.CsvDriver");
	      Connection conCSV = DriverManager.getConnection("jdbc:relique:csv:" + "../data/HITresults/");
	      Statement stmt = conCSV.createStatement();
	      // creating and sending the query
	      String queryCSV = "SELECT * FROM " + csvFile;
	      ResultSet rs = stmt.executeQuery(queryCSV);
	      // using the results
	      while (rs.next())
	      {
	    	  List<String> request = new ArrayList<String>();
	    	  for (int i=1; i<this.columnsCSV.length+1; i++)
	    		  request.add(rs.getString(i));
	    	  this.data.add(request.toArray(new String[0]));
	      }
		  // closing everything
		  rs.close();
		  stmt.close();
		  conCSV.close();
	  }
	  catch(Exception e)  { e.printStackTrace(); }
  }

  
  /**
   * Calls Panos Ipeirotis and his students' <a href="http://code.google.com/p/get-another-label/"><tt>get another label
   * </tt></a> to treat the results of the HIT.
   * 
   * <p>This method has three steps:</p>
   * <ol>
   * 	<li>All the "files" (here, String formatted like files) are generated using the appropriate method, except the
   *		<tt>costFile</tt> which is hard coded.</li>
   *	<li>An ipeirotis.gal.scripts.DawidSkene instance is then created if their is enough HIT results to better their
   *		quality ; otherwise results are taken directly from the <tt>correctFile</tt> String. Indeed, to few data
   *		causes <tt>get another label</tt> to violently crash.</li>
   *	<li>The results are parsed and put in the <tt>data</tt> attribute.</li>
   * </ol>
   * <p>For more details on how <tt>get another label</tt> works, feel free to read its partially commented source files
   * or to browse its absence of documentation.</p>
   */
  public void getanotherlabel(  )
  {
	  // step 1: creation of the "files"
	  String inputFile = generateInputFile() ;
	  String correctFile = generateCorrectFile() ;
	  String costFile = "1\t1\t0\n2\t2\t0\n1\t2\t1\n2\t1\t1" ;
	  Integer iterations = 10 ;
	  HashMap<String,String> posterior_voting = new HashMap <String,String>() ;
	  List<String[]> newData = new ArrayList <String[]> () ;
	  
	  // step 2: get another label is called if their is enough HIT results (here, 10).
	  if (this.data.size() > 10)
		  try
		  {	
			  // the content of this block is roughly a copy-pasta from "get another label" main method.
			  String[] lines_input = inputFile.split("\n");
			  Vector<Labeling> labelings = DawidSkene.loadLabels(lines_input);
			  String[] lines_correct = correctFile.split("\n");
			  Vector<Labeling> correct = DawidSkene.loadLabels(lines_correct);
			  String[] lines_cost = costFile.split("\n") ; 
			  Vector<Labeling> costs = DawidSkene.loadLabels(lines_cost);
			  DawidSkene ds = new DawidSkene(labelings, correct, costs);
			  ds.estimate(iterations);
			  ds.updateAnnotatorCosts();
			  posterior_voting = ds.getMajorityVote();
		  }
	  	  catch (Exception e) { e.printStackTrace(); }

	  // step 2 (again): if their is not enough HIT results, raw results from the "correctFile" String are used directly.
	  else
	  {
		  for (int i=0; i<this.data.size(); i++)
		  {
			  String[] dataLine = this.data.get(i);
			  String sig = signature(dataLine[this.columnAxis], dataLine[this.columnMedia1], dataLine[this.columnMedia2]) ;
			  if (  this.compScores.get(sig) > 0)
				  posterior_voting.put(sig, "1");
			  else
				  posterior_voting.put(sig, "2");
		  }	  
	  }
	  
	  // Step 3: The refined results are put in the "data" attribute
	  for (int i=0; i<this.data.size(); i++)
	  {
		  List<String> newEntry = new ArrayList <String> () ;
		  String sigComparisonStudied = signature( this.data.get(i)[this.columnAxis] , 
				  	this.data.get(i)[this.columnMedia1] , this.data.get(i)[this.columnMedia2]);
		  newEntry.add(this.data.get(i)[this.columnAxis]);
		  newEntry.add(this.data.get(i)[this.columnMedia1]);
		  newEntry.add(this.data.get(i)[this.columnMedia2]);
		  newEntry.add ( posterior_voting.get(sigComparisonStudied) ) ;
		  newData.add(newEntry.toArray(new String[0]));
	  }
	  this.data = newData ;  
  }


  
  /**
   * Creates the "correct file" used by <a href="http://code.google.com/p/get-another-label/"><tt>get another label
   * </tt></a> (it is returned as a String).
   * 
   * <p>"Get another label" needs a "correct file" in its input. For more details, see
   * <a href="http://code.google.com/p/get-another-label/wiki/HowToRun">this</a> (<em>correct file</em> section).</p>
   */
  private String generateCorrectFile()
  {
	  String correctFile = "" ;
	  try
	  {
		  String line;
		  for (int i=0; i<this.data.size(); i++)
		  {
			  String[] dataLine = this.data.get(i);
			  String sig = signature(dataLine[this.columnAxis], dataLine[this.columnMedia1], dataLine[this.columnMedia2]) ;
			  line = dataLine[this.columnWorkerId] + "\t" + sig + "\t" ;
			  if (  this.compScores.get(sig) > 0)
				  line += "1";
			  else
				  line += "2";
			  correctFile += line+"\n";  
		  }
	  }
	  catch (Exception e) { e.printStackTrace(); }
	  return correctFile ;
  }


  /**
   * Creates the "input file" used by 'get another label' (it is returned as a String) and puts the comparisons in
   * the compScores and compNumber Map.
   * 
   * <p>"Get another label" needs an "input file"  in its input. For more details, see
   * <a href="http://code.google.com/p/get-another-label/wiki/HowToRun">this</a> (<em>input file</em> section).</p>
   * 
   * <p>This String formatted like a file (it contains "\n" at the end of each "line") contains the hard results of the
   * HITs. The String is actually generated in the end of the "try" block. In this String, results are recorded they way
   * they are in the csv file, i.e "1" means that the greater of the two media is the first, "2" meaning the contrary.</p>
   * 
   * <p>However, the score of each comparisons is calculated in this function for better performance. The score is an
   * integer. This score is 0 in the beginning, then 1 is added if media1 IS greater than media2 and -1 is if it is
   * actually the contrary. All the hard results are checked only once, results being added to the <tt>compScores</tt> Map
   * attribute as they are checked.</p>
   */
  private String generateInputFile()
  {
	  String inputFile = "";
	  this.compScores = new Hashtable <String,Integer> () ;
	  try{
		  String line;
		  for (int i=0; i<this.data.size(); i++)
		  {
			  String[] dataLine = this.data.get(i);
			  String sig = signature(dataLine[this.columnAxis], dataLine[this.columnMedia1], dataLine[this.columnMedia2]) ;
			
			  // storing the results for future retrieval
			  if (dataLine[this.columnAnswer].equals("Media 1"))
				  if (this.compScores.containsKey(sig))
					  this.compScores.put(sig, this.compScores.get(sig) + 1) ;
				  else
					  this.compScores.put(sig, 1);
			  else
				  if (this.compScores.containsKey(sig))
					  this.compScores.put(sig, this.compScores.get(sig) - 1) ;
				  else
					  this.compScores.put(sig, -1);
			  
			  // writing the "input file"
			  line = dataLine[this.columnWorkerId] + "\t" + sig + "\t" ;
			  if (dataLine[this.columnAnswer].equals("Media 1"))
				  line += "1";
			  else
				  line += "2";
			  inputFile += line + "\n";
		  }
	  }
	  catch (Exception e) { e.printStackTrace(); }
	  return inputFile ;
  }


  /**
   * Returns the <tt>data</tt> attribute.
   * @return	this.data
   */
  public List<String[]> getData()
  {
	  return this.data ;
  }
}
