package myDataBases;


import java.util.*;
// file reading classes
import java.io.File;
// xml reading classes
import org.jdom.*;
import org.jdom.input.SAXBuilder;


/**
 * Allows to retrieve the names of the media to sort as well as other characteristics specified in the
 * <tt>media.xml</tt> file.
 * 
 * <p>The media's names and other characteristics are stored in the <tt>'../data/media.xml'</tt> file. The aim of this
 * class is to allow other parts of this program to access quickly and simply to these informations.</p>
 * 
 * <p>The structure of the media.xml file must be :</p>
 * <p>< content >< media >< id >identifier< /id >< name >the name of the file as used by crowdFlower< /name >
 * < characteristic >a first characteristic< /characteristic >< othercharac >another characteristic< /othercharac >
 * < /media > ... < /content ></p>
 * 
 * <p><strong>IMPORTANT:</strong> Identifiers must go from zero to $number_of_media without any gap. For instance,
 * <tt>[0,2,1,5,4,3]</tt> works (the order does not matter), but [0,1,1337,42,5] does NOT.</p>
 * 
 * 
 * @author	Leo Perrin (perrin.leo@gmail.com)
 */
public class MediaBase {

	//
	// Fields
	//

	/**
	 * The list of the media contained in the file and their characteristics.
	 */
	private Map<Integer, Map<String,String> > content;
	
	//
	// Constructors
	//


	/**
	 * Creates an instance of MediaBase and sets its attributes.
	 * 
	 * <p>Generates a hashMap containing all the media and their characteristics, <tt>content</tt>, by parsing the
	 * <tt>path</tt> xml file. All the information contained in this file will be in the <tt>content</tt> attribute (as
	 * long as it is correctly xml formatted).</p>
	 * 
	 * @param        path Path to the file containing the data
	 */
	public MediaBase( String path )
	{
		this.content = new Hashtable <Integer, Map <String,String>>() ;
		SAXBuilder sxb = new SAXBuilder();
		org.jdom.Document mediaDocument  = null;
		try
		{
			mediaDocument = sxb.build(new File(path));
		}
		catch(Exception e){ e.printStackTrace(); }
		Element rootMedia = mediaDocument.getRootElement();
		Iterator it = rootMedia.getChildren().iterator();
		String field;
		Integer id = 0;
		while (it.hasNext())
		{
			Map<String,String> media = new Hashtable <String,String>();
			Element current = (Element)it.next();
			Iterator jt = current.getChildren().iterator();
			while (jt.hasNext())
			{
				Element attribute = (Element)jt.next();
				field = attribute.getName();
				if (field == "id")
					id = Integer.parseInt( attribute.getValue() );
				else
					media.put(field, attribute.getValue() );
			}
			this.content.put(id,media);
		}
	}

	
	//
	// Methods
	//

	/**
	 * Returns a List of all the identifier of the media considered.
	 * 
	 * <p>Reads the <tt>content</tt> attribute and adds every <tt>< id ></tt> it founds in a list.</p>
	 * 
	 * @return       The list of the media.
	 */
	public List<Integer[]> getAllMedia(  )
	{
		List<Integer[]> result = new ArrayList<Integer[]>();
		List<Integer> array = new ArrayList<Integer>();
		for(Integer i :  this.content.keySet())
			array.add(i);
		result.add(array.toArray(new Integer[0]));
		return result;
	}

	
	/**
	 * Returns <tt>content</tt>, the map of map containing all the information available in the media.xml file.
	 * 
	 * @return content
	 */
	public Map<Integer, Map<String,String> > getContent()
	{ return this.content;  }


	/**
	 * Returns the name of the media wanted.
	 * 
	 * @param        id The identifier of the media wanted.
	 * 
	 * @return       A String, the name of the media whose identifier is <tt>id</tt>.
	 */
	public String getMedia( Integer id )
	{ return this.content.get(id).get("name"); }

}
