package myDataBases;


import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.Statement;
import java.util.*;

import myShell.Interpreter;
import myShell.LogWriter;



/**
 * An instance of this class is used to avoid submitting unnecessary comparisons by saving former results as well as to
 * allow a fast retrieval of these comparison results.
 * 
 * <p>This class is used to store the HITs hard results (i.e, not refined) as they were submitted by the turkers in the
 * <tt>hit</tt> table of the database. Comparisons stored in this table can be false since they are not treated in any way
 * by this program yet. However, a lot of information is available, for instance the country turkers live in, etc. For each
 * comparison between two media, several entries are put in the <tt>hit</tt> table corresponding to the several answers
 * that were made by different turkers.</p>
 * <p>Such hard data is made of several fields ; these fields must be given when an instance of this object is build.</p>
 * <p>MySQLbase is also used to store <tt>refined</tt> results, i.e results obtained using an instance of
 * {@link crowdUser.Refiner}. These results are stored in the <tt>results</tt> table of the database. Each comparison is
 * present only one time : results being refined, they are no longer duplicated.</p>
 * <p>At last, this class can also generate gold data for you to upload at CrowdFLower in order to increase the quality
 * of your results.</p>
 * 
 * @author	Leo Perrin (perrin.leo@gmail.com)
 */
public class MySQLbase {

  //
  // Fields
  //


  /**
   * The url where the database used is.
   */
  private String adr;
  /**
   * A login enabling reading an writing access to the DBcps database.
   */
  private String login;
  /**
   * The password matching the login used.
   */
  private String passwd;
  /**
   * The connection to the MySQL database
   */
  private Connection conSQL;
  /**
   * The labels of the columns in the <tt>hit</tt> table of the MySQL database
   */

// concerning hard data
  private String[] columnsHardDataSQL;
  /**
   * Contains <tt>true</tt> if the content of the corresponding label in the <tt>hit</tt> table is a string and therefore
   * requires quotation marks. 
   */
  private Boolean[] isStringHardData;
  /**
   * The stub common to all queries inserting new hard data in the <tt>hit</tt> table of the SQL database
   */
  private String queryHardDataStub;
  /**
   * The labels of the columns in the <tt>results</tt> table of the MySQL database
   */

// concerning refined results 
  private String[] columnsResultsSQL = { "axis", "idMedia1", "idMedia2", "greater", "nameMedia1", "nameMedia2" };
  /**
   * The same as <tt>isStringHardData</tt>, but for the <tt>results</tt> table. 
   */
  private Boolean[] isStringResults = { true, false, false, false, true, true};
  /**
   * The stub common to all queries inserting results in the <tt>results</tt> table of the SQL database
   */
  private String queryResultsStub ;
  
  
  //
  // Constructors
  //

  /**
   * Creates an instance of the MySQLbase class and sets all of its attributes.
   * 
   * <p>The address of the database, the login of the user able to use the SQL database and its password must be given
   * as parameters. The fields used to store the hard results of the HIT in the <tt>hit</tt> table, as well as those
   * used to store the refined comparisons results in the <tt>results</tt> must also be given.</p>
   * 
   * @param     adr The address of the MySQL database. It is the value to give to the attribute <tt>adr</tt>.
   * @param     login The login of the database. It is the value to give to the attribute <tt>login</tt>.
   * @param     passwd The password of the database. It is the value to give to the attribute <tt>psswd</tt>.
   * @param		fieldsHardData The fields of the <tt>hit</tt> table.
   * @param		fieldsType The types of these fields.
   */
  public MySQLbase( String adr, String login, String passwd, List<String> fieldsHardData,
		  List<String> fieldsType) 
  {
	  this.adr = adr;
	  this.login = login;
	  this.passwd = passwd;
	  this.columnsHardDataSQL = fieldsHardData.toArray(new String[0]) ;
	  
	  // Creation of the isStringHarData array
	  List<Boolean> isFieldAString = new ArrayList <Boolean>() ; 
	  for (String type : fieldsType)
	  {
		  if (type.equals("INT") || type.equals("FLOAT(23)") || type.equals("BIGINT") )
			  isFieldAString.add(false);
		  else
			  isFieldAString.add(true);
		  this.isStringHardData = isFieldAString.toArray(new Boolean[0]);
	  }
	  
	  // Creation of the Hard Data query stub
	  this.queryHardDataStub = "INSERT INTO hit (";
	  for (int i=0; i<this.columnsHardDataSQL.length-1; i++)
		  this.queryHardDataStub += this.columnsHardDataSQL[i]+", ";
	  this.queryHardDataStub += this.columnsHardDataSQL [this.columnsHardDataSQL.length - 1] + ") VALUES (";
	  
	  // Creation of the Results query stub
	  this.queryResultsStub = "INSERT INTO results (";
	  for (int i=0; i<this.columnsResultsSQL.length-1; i++)
		  this.queryResultsStub += this.columnsResultsSQL[i]+", ";
	  this.queryResultsStub += this.columnsResultsSQL [this.columnsResultsSQL.length - 1] + ") VALUES (";
	  
	  // connection to the MySQL base
	  try
	  {
	      Class.forName("com.mysql.jdbc.Driver");	//Register the JDBC driver for MySQL.
	      this.conSQL = DriverManager.getConnection(this.adr,this.login, this.passwd);
	  }
	  catch(Exception e)  { e.printStackTrace(); }
  }
  
  
  //
  // Methods
  //


  /**
   * Tests if the connection with the database works.
   * 
   * <p>This methods allows to test whether the connection with the MySQL database works or not by creating a DriverManager
   * instance connected to the database. The result of this test is saved using the <tt>log</tt> {@link myShell.LogWriter}
   * instance.</p>
   * 
   * @param		log The <tt>LogWriter</tt> instance used to inform the user.
   * @param     testPasswd The password corresponding to the login.
   * 
   * @return	<tt>true</tt> if the connection works, <tt>false</tt> otherwise.
   */
  public static Boolean testConnection(LogWriter log, String testPasswd)
  {
	  Boolean connected = false;
	  Map<String,String> config = Interpreter.getConfig();
	  try
	  {
	      Class.forName("com.mysql.jdbc.Driver");	//Register the JDBC driver for MySQL.
	      DriverManager.getConnection( "jdbc:mysql://localhost:3306/" + config.get("sqlBase") ,
	    		  config.get("sqlUser"), testPasswd);
	      connected = true;
	  }
	  catch(Exception e)
	  {
		  e.printStackTrace();
		  connected = false;
	  }
	  if (connected)
			log.append("connection with the database works.\n");
	  else
			log.append("ERROR : A problem was encountered while connecting with the database.\n" +
			" Interrupting.");
	  return connected;
  }

  /**
   * Returns <tt>true</tt> if the result of the comparison between the media whose IDs are <tt>idMedia1</tt> and
   * <tt>idMedia2</tt> along the <tt>axis</tt> axis is already in the database, <tt>false</tt> otherwise.
   * 
   * <p>This method checks if the comparison between <tt>idMedia1</tt> and <tt>idMedia2</tt> is stored in the
   * <tt>results</tt> table of the database using SQL requests. The roles of <tt>idMedia1</tt> and <tt>idMedia2</tt>
   * ARE the same here, i.e if no comparison <tt>(idMedia1>idMedia2)</tt> were found, a request is sent to see if
   * <tt>(idMedia2>idMedia1)</tt> is present.</p>
   *  
   * @param     idMedia1	The identifier of the first media being compared.
   * @param     idMedia2	The identifier of the second media to be compared.
   * @param		axis 		The axis along which the comparison is made.
   * 
   * @return    <tt>true</tt> if the comparison is already known, <tt>false</tt> otherwise.
   */
  public boolean alreadyDone( int idMedia1, int idMedia2, String axis )
  {
	  String query1 = "SELECT greater FROM results WHERE axis='"+axis;
	  String query2 = "SELECT greater FROM results WHERE axis='"+axis;
	  boolean result = false;
	  
	  query1 += "' AND idMedia1="+idMedia1+" AND idMedia2="+idMedia2+" ;";
	  query2 += "' AND idMedia1="+idMedia2+" AND idMedia2="+idMedia1+" ;";
	  
	  try
	  {
		  Statement stmt = this.conSQL.createStatement(ResultSet.TYPE_SCROLL_INSENSITIVE,ResultSet.CONCUR_READ_ONLY);
		  ResultSet rs;
		  rs = stmt.executeQuery(query1);
		  if (rs.next())
			  result = true;
		  else
		  {
			  rs = stmt.executeQuery(query2);
			  if (rs.next())
				  result = true;
		  }
		  rs.close();
		  stmt.close();
	  }
	  catch(Exception e)  { e.printStackTrace(); }
	  return result;
  }


  /**
   * Returns the result of the comparison between the two media as stored in the SQL database.
   * 
   * <p>Calling this method requires <tt>alreadyPermformed(idMedia1,idMedia2,axis)</tt> to be true because such a
   * verification is NOT performed here (due to performance considerations).</p>
   * 
   * @param     idMedia1	The identifier of the first media being compared.
   * @param     idMedia2	The identifier of the second media to be compared.
   * @param		axis 		The axis along which the comparison is made.
   * 
   * @return    true if <tt>idMedia1</tt> has been tagged as bigger than <tt>idMedia2</tt>, false otherwise.
   */
  public Boolean comparisonResult( int idMedia1, int idMedia2 , String axis)
  {
	  String query1 = "SELECT greater FROM results WHERE axis='"+axis;
	  String query2 = "SELECT greater FROM results WHERE axis='"+axis;
	  boolean result = true;
	  
	  query1 += "' AND idMedia1="+idMedia1+" AND idMedia2="+idMedia2+" ;";
	  query2 += "' AND idMedia1="+idMedia2+" AND idMedia2="+idMedia1+" ;";
	  
	  try
	  {
		  Statement stmt = this.conSQL.createStatement(ResultSet.TYPE_SCROLL_INSENSITIVE,ResultSet.CONCUR_READ_ONLY);
		  ResultSet rs;
		  rs = stmt.executeQuery(query1);
		  if (rs.next())
			  result = (rs.getInt("greater") == 1 );
		  else
		  {
			  rs = stmt.executeQuery(query2);
			  if (rs.next())
				  result = (rs.getInt("greater") == 2 );
		  }
		  rs.close();
		  stmt.close();
	  }
	  catch(Exception e)  { e.printStackTrace(); }
	  return result;
  }


  
  /**
   * Inserts the content of <tt>entries</tt> in the <tt>hit</tt> table of the MySQL database.
   * 
   * <p><tt>entries'</tt> arrays' content must be in the correct order so as for each column to have the correct "name",
   * i.e for example, if the first column of <tt>this.columnsHardData</tt> is "cfId", the first element of each array
   * contained in <tt>entries</tt> must be the "cfId" of a HIT.</p>
   * <p>This method builds SQL request using the content of <tt>entries</tt>, but also that of <tt>columnsHardDataSQL</tt>
   * and <tt>isStringHardData</tt>. The list of array is went through within a loop; in which another loop goes through
   * each element of the array. For each one, it is checked whether it is string or not: if it is, quotation marks are
   * added to the request (this is what <tt>isStringHardData</tt> is for).</p>
   *   
   * @param		entries	A List of array of String, each array being the result of one HIT.
   */
  public void insertHardDataInSQL( List<String[]> entries)
  {  
	  for (int i=0; i< entries.size(); i++)
	  {
		  String[] entry = entries.get(i);
		  String query = this.queryHardDataStub;
		  Integer endLoop;
		  
		  // Determining the number of fields to write in the SQL request
		  if (entry.length < this.columnsHardDataSQL.length )
			  endLoop = entry.length;
		  else
			  endLoop = this.columnsHardDataSQL.length;
		  
		// Writing each field but the last one
  		  for (int j=0; j<endLoop; j++)
  		  {
  			  if (this.isStringHardData[j])	// if it is a string, we print it between "'"
  				  query += "'" + entry[j] + "'";
  			  else	// if it is a number
  				  if (entry[j].length() == 0)	// if it is null, we write 0
  					  query += "0";
  				  else	// else we write it
  					  query += entry[j];
  			  if (j == endLoop-1)	// are we at the end of the request ?
  				  query += ") ;\n";	// if yes, we close it with a parenthesis and a semicolon
  			  else
  				  query += ", ";	// otherwise, we just a put a comma
  		  }
		  try
		  {
			  Statement stmt = this.conSQL.createStatement();
			  stmt.executeUpdate(query);
		  }
		  catch(Exception e)  { e.printStackTrace(); }
	  }
  }


  
  /**
   * Inserts the content of <tt>entries</tt> in the <tt>results</tt> table of the MySQL database.
   * 
   * <p><tt>entries</tt> must have the correct structure, it must fit with that of the <tt>columnsResults</tt> attribute.
   * For instance, the first element of each array of String must be the <tt>axis</tt> along which the comparison was
   * performed.</p>
   * <p>Actually, you should <strong>not</strong> <a href="http://www.youtube.com/watch?v=otCpCn0l4Wo">touch this</a>,
   * unless you are making heavy modification.</p>
   * 
   * @param		entries A list containing the results after their treatment.
   * @param		dbNames	The <tt>MediaBase</tt> instance in which the names of the media are stored.
   */
  public void insertResultsInSQL( List<String[]> entries, MediaBase dbNames)
  {  
	  for (int i=0; i<entries.size(); i++)
	  {
		  String[] entry = entries.get(i);
		  String query = this.queryResultsStub;
		  // adding the content of the entry in the query
		  for (int j=0; j<entry.length-1; j++)
			  if (this.isStringResults[j])
				  query += "'" + entry[j] + "', ";
			  else
				  query += entry[j] + ", ";
		  // adding the "greater" entry
		  query += entry[3] + ", " ;
		  // adding the names of the media considered
		  query += "'" + dbNames.getMedia( Integer.parseInt(entry[1]) ) + "', '"
		  			+ dbNames.getMedia( Integer.parseInt(entry[2]) )  + "') ;" ;
		  try
		  {
			  //System.out.println(query);
			  Statement stmt = this.conSQL.createStatement();
			  stmt.executeUpdate(query);
		  } catch(Exception e)  { e.printStackTrace(); }
	  }
  }
  
  
  

  /**
   * Creates a dump of the database.
   * 
   * <p>A file named <tt>DBcps_dd_mm_yy.sql</tt> is written in the <tt>data/mysql/dump</tt> folder. It contains all the
   * information necessary to re-create a MySQL database containing all the comparisons. In case of failure of the CPS,
   * this dump enables the process to start again where the failure occurred without requiring submitting costly HITs
   * again.</p>
   */
  public void generateDump()
  {
	  try
	  {
		  Calendar cal = Calendar.getInstance();
		  String dumpName = "DBcps_" + cal.get(Calendar.DATE) + "-" + cal.get(Calendar.MONTH)
		  					+ "-" + cal.get(Calendar.YEAR) + ".sql";
		  String [] cmd = { "../data/mysql/dump.sh",this.login,this.passwd, dumpName};
		  Runtime.getRuntime().exec(cmd);
	  }
	  catch(Exception e)  { e.printStackTrace(); }
  }

}
